//#include "I2Cdev.h"
//#include "MPU6050.h"
#include <Wire.h>
#include "SparkFun_BNO080_Arduino_Library.h"
#include "utility/imumaths.h"
#include <math.h>
#include "soc/soc.h"
#include "soc/rtc_cntl_reg.h"
#include "sdkconfig.h"

// INITIALIZE BNO080 
  BNO080 bno080;
  #define BUF_LEN 20
  byte byteBuffer[BUF_LEN];
  char charBuffer[BUF_LEN] = {0x00};
  //String deviceName = "rwaht00";

// battery measure initialize
//  const int BAT_PIN = A13; // Messure half the battery voltage
//  float batVoltage = 0;  

// for millis() time management from jasch instead of delay:
uint32_t now;
uint32_t prev;
uint32_t slowtime;
uint32_t now2;
uint32_t now3;
uint32_t delta_time;
uint32_t delta_time_intra;

float yaw_degrees_f, pitch_degrees_f, roll_degrees_f, lin_accel_x_f, lin_accel_y_f, lin_accel_z_f;

// SET MAX ACCELERATION (+/-) here:
float accel_max = 50.0;

// the 6 values of the MPU
//int ax, ay, az;
//int gx, gy, gz;
unsigned int ax, ay, az;
unsigned int gx, gy, gz;

// bit startpoint for parsing the serial data
uint8_t start = 255;

// MSB/LSB variables for encapsulation
uint8_t msbax, lsbax, msbay, lsbay, msbaz, lsbaz;
uint8_t msbgx, lsbgx, msbgy, lsbgy, msbgz, lsbgz;

void setup()
{
  // begin Software Serial Port
  Serial.begin(38400);
  Serial.println("start of setup");
    
  // join I2C bus (I2Cdev library doesn't do this automatically)
  // #if I2CDEV_IMPLEMENTATION == I2CDEV_ARDUINO_WIRE
  //Wire.begin();
  // #elif I2CDEV_IMPLEMENTATION == I2CDEV_BUILTIN_FASTWIRE
  // Fastwire::setup(400, true);
  // #endif

  delay(500);
  
  Wire.flush();
  
  // initialize device
  bno080.begin();

  //Wire.begin(23,22);
  Wire.begin();
  Wire.setClock(400000); //Increase I2C data rate to 400kHz

  bno080.enableRotationVector(20);       // timeBetweenReports = 20 ms 
  bno080.enableLinearAccelerometer(20);  // timeBetweenReports = 20 ms 
      // ???????? what is >linear< accelerometer?

  // gibt auch noch:
  // .enableGameRotationVector(20);
  // .enableGyro(20);
  // .enableMagnetometer(20);
  // .enableStepCounter(20);
  // etc.

  // und "Calibration Commands to sensor" (?):
  // .sendCalibrateCommand(CALIBRATE_ACCEL);
  // .sendCalibrateCommand(CALIBRATE_GYRO);
  // .sendCalibrateCommand(CALIBRATE_MAG);
  // .sendCalibrateCommand(CALIBRATE_PLANAR_ACCEL);
  // .sendCalibrateCommand(CALIBRATE_ACCEL_GYRO_MAG); // all sensors
  // .sendCalibrateCommand(CALIBRATE_STOP);

  delay(500);

  // MAYBE DO THIS??? (from jasch script)
 // bno.setExtCrystalUse(true);

  // from millis() timing from jasch instead of delay in loop()
    slowtime = millis();
    prev = millis();
    now3 = micros(); // prev time in microsecs

  //batVoltage = 0.002 * analogRead(BAT_PIN);
  //Serial.print("Battery: ");
  //Serial.print(batVoltage);
  //Serial.println(" V");
}

void loop()
{
  now = millis();
  delta_time = now - prev;
  if (delta_time >= 10) // read / write every 10 ms -> 100 Hz
  {
    now2 = micros(); // current time in microsecs for measuring read-time

  // * * * * * * * * * * * * * * * * * * * * 
  
  // READ MEASUREMENTS FROM BNO080

  //Serial.print("Data from BNO080 available: ");
  //Serial.println(bno080.dataAvailable());
  
  if (bno080.dataAvailable() == true) {   // checks if NEW data is available from the imu
        //Serial.println("Debug in");
        float quatI = bno080.getQuatI();  // retrieves i-axis quaternion from imu
        float quatJ = bno080.getQuatJ();  // retrieves j-axis...
        float quatK = bno080.getQuatK();  // retrieves k-axis...
        float quatReal = bno080.getQuatReal(); //retrieves the real component of the quaternion from the imu

        imu::Quaternion quat = imu::Quaternion(quatReal, quatI, quatJ, quatK);
        quat.normalize();
        imu::Vector<3> q_to_euler = quat.toEuler();
            
      float yaw = q_to_euler.x();
      yaw_degrees_f = yaw * 180.0 / M_PI; // conversion to degrees
      if( yaw_degrees_f < 0 ) 
        yaw_degrees_f += 360.0; // convert negative to positive angles
      
      float pitch = q_to_euler.y();
      pitch_degrees_f = pitch * 180.0/ M_PI;
      pitch_degrees_f += 90.0;
      if( pitch_degrees_f < 0 ) 
        pitch_degrees_f += 180.0; // convert negative to positive angles

      float roll = q_to_euler.z();
      roll_degrees_f = roll * 180.0 / M_PI;
      roll_degrees_f += 180.0;
      if( roll_degrees_f < 0 )
        roll_degrees_f += 360.0; // convert negative to positive angles

      lin_accel_x_f = 0.0;
      lin_accel_x_f = bno080.getLinAccelX(); // float

      lin_accel_y_f = 0.0;
      lin_accel_y_f = bno080.getLinAccelY(); // float

      lin_accel_z_f = 0.0;
      lin_accel_z_f = bno080.getLinAccelZ(); // float

//    Serial.println();
//    Serial.print("Roll: ");
//    Serial.print(roll_degrees);
//    Serial.print(", Pitch: ");
//    Serial.print(pitch_degrees);
//    Serial.print(", Yaw: ");
//    Serial.print(yaw_degrees);
//    Serial.println();

//    float directRoll = (bno080.getRoll()) *180.0 / PI;
//    float directPitch = (bno080.getPitch()) *180.0 / PI;
//    float directYaw = (bno080.getYaw()) *180.0 / PI;
//    Serial.print("directRoll: ");
//    Serial.print(eRoll);
//    Serial.print(", directPitch: ");
//    Serial.print(ePitch);
//    Serial.print(", directYaw: ");
//    Serial.print(eYaw);
//    Serial.println();

     // unsigned int steps = 0;
     // steps = bno080.getStepCount();
      //step_str = String(steps);
      // Set coordinates
       
      // Data package
//      dataStr = yaw_degrees_str + " " + pitch_degrees_str + " " + lin_accel_z_str + " " + step_str;
      //      // Send data
      //      pCharacteristicTracking->setValue(dataStr.c_str());
      //      // Say here is new data
      //      pCharacteristicTracking->notify();
      
      // Print data
           // Serial.println(dataStr);
           
      //Serial.println("Debug print");

    //float quatRadianAccuracy = bno080.getQuatRadianAccuracy();
    //Serial.print(F("QuatRadianAccuracy: "));
    //Serial.print(quatRadianAccuracy, 2);
    //uint64_t chipid=ESP.getEfuseMac();
    //Serial.printf("\nESP32 Chip ID = %04X",(uint16_t)(chipid>>32));//print High 2 bytes
   // Serial.printf("%08X\n",(uint32_t)chipid);//print Low 4bytes.
    
  }
    
  // * * * * * * * * * * * * * * * * * * * * 



  // * * * * * * * * * * * * * * * * * * * * 

  // convert bno080 to cedric's mpu format:

  // ax = acceleration on x axis
  // gx = gyro on x axis
  // but what i get from bno080 is:
  //    yaw_degrees, pitch_degrees, lin_accel_z_f (float)

    // YAW (360 deg), PITCH (180 deg), ROLL (360 deg):
    gx = (int)(round(yaw_degrees_f * (65535 / 360.0))); // round and cast int
    gy = (int)(round(pitch_degrees_f * (65535 / 180.0))); 
    gz = (int)(round(roll_degrees_f * (65535 / 360.0)));

    // ACCEL VALUES for X, Y, Z - here scaled for values from 0 to +- 50.0:
    ax = (int)(round((lin_accel_x_f + 1.0 * accel_max) * (65535 / (2.0 * accel_max))));
    ay = (int)(round((lin_accel_y_f + 1.0 * accel_max) * (65535 / (2.0 * accel_max))));
    az = (int)(round((lin_accel_z_f + 1.0 * accel_max) * (65535 / (2.0 * accel_max))));
  
  // * * * * * * * * * * * * * * * * * * * * 

  // read raw accel/gyro measurements from device
  // mpu.getMotion6(&ax, &ay, &az, &gx, &gy, &gz);
  
  // convert sensor values from int to uint (0...65535)
 // ax += 32768;
 // ay += 32768;
 // az += 32768;
        
 // gx += 32768;
 // gy += 32768;
 // gz += 32768;

  // cut off highest values (2^16-1, reserved for the "start" bytes)
  
  if (ax == 65535) ax = 65534;
  if (ay == 65535) ay = 65534;
  if (az == 65535) az = 65534;

  if (gx == 65535) gx = 65534;
  if (gy == 65535) gy = 65534;
  if (gz == 65535) gz = 65534;

//  ax = 65534;
//  ay = 65534;
//  az = 65534;
//  gx = 65534;
//  gy = 65534;
//  gz = 65534;

  // send int16_t as 2 int8_ts for transmission and inhibit consecutive 255 bytes:
  // every time, two neighbouring bytes of the 12 byte structure (except for the outer ones)
  //   are both 255, reduce the lsb to 254 (lsb, because it doesn't "hurt" the signal)
  msbax = (uint8_t)(ax >> 8); lsbax = (uint8_t)(ax & 255);
  msbay = (uint8_t)(ay >> 8); lsbay = (uint8_t)(ay & 255);
  if(lsbax == 255 && msbay == 255) {
    lsbax = 254;
  }
  msbaz = (uint8_t)(az >> 8); lsbaz = (uint8_t)(az & 255);
  if(lsbay == 255 && msbaz == 255) {
    lsbay = 254;
  }
  msbgx = (uint8_t)(gx >> 8); lsbgx = (uint8_t)(gx & 255);
  if(lsbaz == 255 && msbgx == 255) {
    lsbaz = 254;
  }
  msbgy = (uint8_t)(gy >> 8); lsbgy = (uint8_t)(gy & 255);
  if(lsbgx == 255 && msbgy == 255) {
    lsbgx = 254;
  }
  msbgz = (uint8_t)(gz >> 8); lsbgz = (uint8_t)(gz & 255);
  if(lsbgy == 255 && msbgz == 255) {
    lsbgy = 254;
  }
  if(lsbgz == 255) {
    lsbgz = 254;
  }

  // transmit values
    Serial.write(start); Serial.write(start);
    Serial.write(msbax); Serial.write(lsbax);
    Serial.write(msbay); Serial.write(lsbay);
    Serial.write(msbaz); Serial.write(lsbaz);
    Serial.write(msbgx); Serial.write(lsbgx);
    Serial.write(msbgy); Serial.write(lsbgy);
    Serial.write(msbgz); Serial.write(lsbgz);
    Serial.write(0);
    delay(1);

  // * * * * * * * * * * * * * * * * * * * * 

  // MAYBE CORRECT THIS with MILLIS instead of DELAY (here: 100 Hz = 10ms)
  prev = millis();
  delta_time_intra = now2 - now3;
  now3 = micros();
  }
  // * * * * * * * * * * * * * * * * * * * * 
  
}
